from typing import Dict, Any
import aiohttp

from wcpan.drive.core.exceptions import DriveError, DownloadError


class AuthenticationError(DriveError):
    pass


class CredentialFileError(DriveError):

    @staticmethod
    def _get_dummy_config_format() -> str:
        import json
        dict_ = {
            'installed': {
                'redirect_uris': [
                    '__URI__',
                ],
                'client_id': '__ID__',
                'client_secret': '__SECRET__',
            },
        }
        return json.dumps(dict_, ensure_ascii=False, indent=2)

    def __init__(self) -> None:
        super().__init__(self, (
            'invalid config file format, correct example:\n'
            f'{CredentialFileError._get_dummy_config_format()}'
        ))


class TokenFileError(DriveError):
    pass


class NetworkError(DriveError):
    pass


class ResponseError(DriveError):

    def __init__(self,
        status: str,
        response: aiohttp.ClientResponse,
        json_: Dict[str, Any],
    ) -> None:
        self._status = status
        self._response = response
        self._message = f'{self.status} {self._response.reason} - {json_}'
        self._json = json_

    def __str__(self) -> str:
        return self._message

    @property
    def status(self) -> str:
        return self._status

    @property
    def json(self) -> Any:
        return self._json


class DownloadAbusiveFileError(DownloadError):
    pass


class InvalidAbuseFlagError(DownloadError):
    pass


class InvalidRangeError(DownloadError):
    pass



import asyncio
import contextlib
import enum
import math
import random
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Callable,
    Dict,
    Generator,
    Optional,
    Tuple,
    Union,
)

import aiohttp
from wcpan.logger import DEBUG, EXCEPTION, WARNING

from .util import OAuth2Storage, OAuth2Manager
from .exceptions import (
    DownloadAbusiveFileError,
    InvalidAbuseFlagError,
    InvalidRangeError,
    NetworkError,
    ResponseError,
)


BACKOFF_FACTOR = 2
BACKOFF_STATUSES = set(('403', '408', '429', '500', '502', '503', '504'))

ContentProducer = Callable[[], AsyncGenerator[bytes, None]]
ReadableContent = Union[bytes, ContentProducer]


class Network(object):

    def __init__(self, storage: OAuth2Storage, timeout: int) -> None:
        self._storage = storage
        self._timeout = timeout
        self._backoff_level = 0
        self._session: aiohttp.ClientSession = None
        self._oauth: OAuth2Manager = None
        self._raii = None

    async def __aenter__(self) -> 'Network':
        async with contextlib.AsyncExitStack() as stack:
            self._session = await stack.enter_async_context(
                aiohttp.ClientSession())
            self._oauth = await stack.enter_async_context(
                OAuth2Manager(self._session, self._storage))
            self._raii = stack.pop_all()

        return self

    async def __aexit__(self, type_, value, traceback) -> bool:
        await self._raii.aclose()
        self._backoff_level = 0
        self._session = None
        self._oauth = None
        self._raii = None

    async def fetch(self,
        method: str,
        url: str,
        args: Dict[str, Any] = None,
        headers: Dict[str, str] = None,
        body: ReadableContent = None,
    ) -> 'JSONResponse':
        while True:
            kwargs = await self._prepare_kwargs(method, url, args, headers,
                                                body)

            try:
                response = await self._request_loop(kwargs)
            except aiohttp.ClientConnectionError as e:
                continue

            try:
                return await to_json_response(response)
            except aiohttp.ClientConnectionError as e:
                # The server recived the request, but reading body failed.
                # We are in a broken state, just let client to handle it.
                raise NetworkError() from e

    async def upload(self,
        method: str,
        url: str,
        args: Dict[str, Any] = None,
        headers: Dict[str, str] = None,
        body: ReadableContent = None,
    ) -> 'JSONResponse':
        kwargs = await self._prepare_kwargs(method, url, args, headers, body)
        # NOTE Upload can take long time to send data, so we only set timeout
        # for socket connection.
        kwargs['timeout'] = aiohttp.ClientTimeout(sock_connect=self._timeout)

        try:
            response = await self._request_loop(kwargs)
        except aiohttp.ClientConnectionError as e:
            raise NetworkError() from e

        return await to_json_response(response)

    async def download(self,
        method: str,
        url: str,
        args: Dict[str, Any] = None,
        headers: Dict[str, str] = None,
        body: ReadableContent = None,
    ) -> 'StreamResponse':
        while True:
            kwargs = await self._prepare_kwargs(method, url, args, headers,
                                                body)
            # NOTE Download can take long time to send data, so we only set
            # timeout for socket connection.
            kwargs['timeout'] = aiohttp.ClientTimeout(
                sock_connect=self._timeout)

            try:
                response = await self._request_loop(kwargs)
            except aiohttp.ClientConnectionError:
                continue

            return StreamResponse(response, self._timeout)

    async def _request_loop(self,
        kwargs: Dict[str, Any],
    ) -> aiohttp.ClientResponse:
        while True:
            await self._wait_backoff()

            try:
                assert self._session is not None
                response = await self._session.request(**kwargs)
            except aiohttp.ClientConnectionError:
                self._adjust_backoff_level(True)
                raise

            status = str(response.status)
            rv = await self._check_status(status, response)
            if rv == Status.OK:
                return response
            if rv == Status.REFRESH:
                assert self._oauth is not None
                await self._oauth.renew_token()
                await self._update_token_header(kwargs['headers'])
                continue
            if rv == Status.BACKOFF:
                continue

            try:
                json_ = await response.json()
            except aiohttp.ContentTypeError as e:
                text = await response.text()
                EXCEPTION('wcpan.drive.google', e) << text
                raise

            self._raiseError(status, response, json_)

    async def _prepare_kwargs(self,
        method: str,
        url: str,
        args: Optional[Dict[str, Any]],
        headers: Optional[Dict[str, str]],
        body: Optional[ReadableContent],
    ) -> Dict[str, Any]:
        kwargs = {
            'method': method,
            'url': url,
            'headers': await self._prepare_headers(headers),
        }
        if args is not None:
            kwargs['params'] = list(normalize_query_string(args))
        if body is not None:
            kwargs['data'] = body if not callable(body) else body()
        return kwargs

    async def _prepare_headers(self,
        headers: Optional[Dict[str, str]],
    ) -> Dict[str, str]:
        if headers is None:
            h = {}
        else:
            h = {k: v if isinstance(v, (bytes, str)) or v is None else str(v)
                 for k, v in headers.items()}
        await self._update_token_header(h)
        return h

    async def _update_token_header(self, headers: Dict[str, str]) -> None:
        assert self._oauth is not None
        token = await self._oauth.get_access_token()
        headers['Authorization'] = f'Bearer {token}'

    async def _check_status(self,
        status: str,
        response: aiohttp.ClientResponse,
    ) -> 'Status':
        backoff = await backoff_needed(status, response)
        self._adjust_backoff_level(backoff)
        if backoff:
            # rate limit error, too many request, server error
            return Status.BACKOFF

        # normal response
        if status[0] in ('1', '2', '3'):
            return Status.OK

        # need to refresh access token
        if status == '401':
            return Status.REFRESH

        # otherwise it is an error
        return Status.UNKNOWN

    def _adjust_backoff_level(self, backoff: bool) -> None:
        if backoff:
            self._backoff_level = min(self._backoff_level + 2, 10)
        else:
            self._backoff_level = max(self._backoff_level - 1, 0)

    async def _wait_backoff(self) -> None:
        if self._backoff_level <= 0:
            return
        seed = random.random()
        power = 2 ** self._backoff_level
        s_delay = math.floor(seed * power * BACKOFF_FACTOR)
        s_delay = min(self._timeout, s_delay)
        DEBUG('wcpan.drive.google') << 'backoff for' << s_delay
        await asyncio.sleep(s_delay)

    def _raiseError(self,
        status: str,
        response: aiohttp.ClientResponse,
        json_: Dict[str, Any],
    ) -> None:
        if status == '403':
            firstError = json_['error']['errors'][0]
            reason = firstError['reason']
            if reason == 'cannotDownloadAbusiveFile':
                raise DownloadAbusiveFileError(firstError['message'])
            if reason == 'invalidAbuseAcknowledgment':
                raise InvalidAbuseFlagError(firstError['message'])
        elif status == '416':
            firstError = json_['error']['errors'][0]
            reason = firstError['reason']
            if reason == 'requestedRangeNotSatisfiable':
                raise InvalidRangeError(response.request_info.headers['Range'])
        raise ResponseError(status, response, json_)


class Request(object):

    def __init__(self, request: aiohttp.RequestInfo) -> None:
        self._request = request

    @property
    def uri(self) -> str:
        return self._request.url

    @property
    def method(self) -> str:
        return self._request.method

    @property
    def headers(self) -> Dict[str, str]:
        return self._request.headers


class Response(object):

    def __init__(self, response: aiohttp.ClientResponse) -> None:
        self._response = response
        self._status = str(response.status)
        self._request = Request(response.request_info)

    @property
    def status(self):
        return self._status

    def get_header(self, key: str) -> str:
        h = self._response.headers.getall(key)
        return None if not h else h[0]


class JSONResponse(Response):

    def __init__(self,
        response: aiohttp.ClientResponse,
        json_: Dict[str, Any],
    ) -> None:
        super().__init__(response)
        self._json = json_

    @property
    def json(self) -> Dict[str, Any]:
        return self._json


class StreamResponse(Response):

    def __init__(self, response: aiohttp.ClientResponse, timeout: int) -> None:
        super().__init__(response)
        self._timeout = timeout

    async def __aenter__(self) -> 'StreamResponse':
        await self._response.__aenter__()
        return self

    async def __aexit__(self, type_, exc, tb) -> bool:
        await self._response.__aexit__(type_, exc, tb)

    async def chunks(self) -> AsyncIterator[bytes]:
        g = self._response.content.iter_any()
        while True:
            try:
                v = await asyncio.wait_for(g.__anext__(), self._timeout)
            except StopAsyncIteration:
                break
            yield v

    async def read(self, length):
        return await self._response.content.read(length)


async def to_json_response(response: aiohttp.ClientResponse) -> JSONResponse:
    async with response:
        if response.content_type == 'application/json':
            json_ = await response.json()
        else:
            json_ = await response.text()
    return JSONResponse(response, json_)


class Status(enum.Enum):

    OK = enum.auto()
    REFRESH = enum.auto()
    BACKOFF = enum.auto()
    UNKNOWN = enum.auto()


async def backoff_needed(
    status: str,
    response: aiohttp.ClientResponse,
) -> bool:
    if status not in BACKOFF_STATUSES:
        return False

    # not all 403 errors are rate limit error
    if status == '403':
        msg = await response.json()
        if not msg:
            # undefined behavior, probably a server problem, better backoff
            WARNING('wcpan.drive.google') << '403 with empty error message'
            return True
        domain = msg['error']['errors'][0]['domain']
        if domain != 'usageLimits':
            return False

    # the request timedout
    if status == '408':
        # NOTE somehow this error shows html instead of json
        WARNING('wcpan.drive.google') << '408 request timed out'
        # No need to backoff because in this case the whole request cannot be
        # resumed at all.
        return False

    return True


def normalize_query_string(
    qs: Dict[str, Any],
) -> Generator[Tuple[str, str], None, None]:
    for key, value in qs.items():
        if isinstance(value, bool):
            value = 'true' if value else 'false'
        elif isinstance(value, (int, float)):
            value = str(value)
        elif not isinstance(value, str):
            raise ValueError('unknown type in query string')
        yield key, value



import asyncio
import contextlib
import hashlib
import re
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
)

from wcpan.logger import INFO, DEBUG, EXCEPTION, WARNING

from wcpan.drive.core.abc import (
    RemoteDriver,
    ReadableFile,
    WritableFile,
    Hasher,
)
from wcpan.drive.core.exceptions import (
    NodeConflictedError,
    InvalidNameError,
    NodeNotFoundError,
    ParentNotFoundError,
    UploadError,
)
from wcpan.drive.core.types import (
    ChangeDict,
    MediaInfo,
    Node,
    NodeDict,
    PrivateDict,
    ReadOnlyContext,
)

from .api import Client
from .exceptions import DownloadAbusiveFileError, ResponseError
from .util import FOLDER_MIME_TYPE, OAuth2Storage


GoogleFileDict = Dict[str, Any]


FILE_FIELDS = ','.join([
    'id',
    'name',
    'mimeType',
    'trashed',
    'parents',
    'createdTime',
    'modifiedTime',
    'md5Checksum',
    'size',
    'shared',
    'ownedByMe',
    'imageMediaMetadata/width',
    'imageMediaMetadata/height',
    'videoMediaMetadata/width',
    'videoMediaMetadata/height',
    'videoMediaMetadata/durationMillis',
    'appProperties',
])
CHANGE_FIELDS = ','.join([
    'nextPageToken',
    'newStartPageToken',
    f'changes(fileId,removed,file({FILE_FIELDS}))',
])
LIST_FIELDS = ','.join([
    'nextPageToken',
    f'files({FILE_FIELDS})',
])


class GoogleDriver(RemoteDriver):

    @classmethod
    def get_version_range(cls):
        return (1, 1)

    def __init__(self, context: ReadOnlyContext) -> None:
        self._storage = OAuth2Storage(context.config_path, context.data_path)
        self._timeout = 60
        self._client = None
        self._raii = None

    async def __aenter__(self) -> RemoteDriver:
        async with contextlib.AsyncExitStack() as stack:
            self._client = await stack.enter_async_context(
                Client(self._storage, self._timeout))
            self._raii = stack.pop_all()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> bool:
        await self._raii.aclose()
        self._client = None
        self._raii = None

    async def get_initial_check_point(self) -> str:
        return '1'

    async def fetch_root_node(self) -> Node:
        rv = await self._client.files.get('root', fields=FILE_FIELDS)
        rv = rv.json
        rv['name'] = None
        rv['parents'] = []
        node = node_from_api(rv)
        return node

    async def fetch_changes(self,
        check_point: str,
    ) -> AsyncGenerator[Tuple[str, List[ChangeDict]], None]:
        new_start_page_token = None
        changes_list_args = {
            'page_token': check_point,
            'page_size': 1000,
            'restrict_to_my_drive': True,
            'fields': CHANGE_FIELDS,
        }

        while new_start_page_token is None:
            rv = await self._client.changes.list_(**changes_list_args)
            rv = rv.json
            next_page_token = rv.get('nextPageToken', None)
            new_start_page_token = rv.get('newStartPageToken', None)
            changes = rv['changes']

            check_point = next_page_token if next_page_token is not None else new_start_page_token
            changes = list(normalize_changes(changes))

            yield check_point, changes

            changes_list_args['page_token'] = check_point

    async def download(self, node: Node) -> ReadableFile:
        rf = GoogleReadableFile(self._client.files.download, node)
        return rf

    async def create_folder(self,
        parent_node: Node,
        folder_name: str,
        private: Optional[PrivateDict],
        exist_ok: bool,
    ) -> Node:
        # do not create again if there is a same file
        node = await self._fetch_node_by_name_from_parent_id(
            folder_name,
            parent_node.id_,
        )
        if node:
            if exist_ok:
                INFO('wcpan.drive.google') << 'skipped (existing)' << folder_name
                return node
            else:
                raise NodeConflictedError(node)

        api = self._client.files
        rv = await api.create_folder(folder_name=folder_name,
                                     parent_id=parent_node.id_,
                                     app_properties=private)
        rv = rv.json
        node = await self._fetch_node_by_id(rv['id'])

        return node

    async def upload(self,
        parent_node: Node,
        file_name: str,
        file_size: Optional[int],
        mime_type: Optional[str],
        media_info: Optional[MediaInfo],
        private: Optional[PrivateDict],
    ) -> WritableFile:
        # do not upload if remote exists a same file
        node = await self._fetch_node_by_name_from_parent_id(
            file_name,
            parent_node.id_,
        )
        if node:
            raise NodeConflictedError(node)

        api = self._client.files
        wf = GoogleWritableFile(
            initiate=api.initiate_uploading,
            upload=api.upload,
            get_status=api.get_upload_status,
            touch=api.create_empty_file,
            get_node=self._fetch_node_by_id,
            parent_id=parent_node.id_,
            name=file_name,
            timeout=self._timeout,
            size=file_size,
            mime_type=mime_type,
            media_info=media_info,
            private=private,
        )
        return wf

    async def trash_node(self, node: Node) -> None:
        await self._client.files.update(node.id_, trashed=True)

    async def rename_node(self,
        node: Node,
        new_parent: Optional[Node],
        new_name: Optional[str],
    ) -> Node:
        fnbnfpi = self._fetch_node_by_name_from_parent_id
        parent_id = node.parent_id if not new_parent else new_parent.id_
        name = node.name if not new_name else new_name
        # make sure it does not conflict to existing node
        new_node = await fnbnfpi(name, parent_id)
        if new_node:
            raise NodeConflictedError(new_node)

        kwargs = {
            'file_id': node.id_,
        }
        if new_name:
            kwargs['name'] = new_name
        if new_parent and new_parent.id_ != node.parent_id:
            kwargs['add_parents'] = [new_parent.id_]
            kwargs['remove_parents'] = [node.parent_id]

        dummy_rv = await self._client.files.update(**kwargs)
        node = await self._fetch_node_by_id(node.id_)
        return node

    async def get_hasher(self) -> Hasher:
        return PicklableHasher()

    async def _fetch_node_by_name_from_parent_id(self,
        name: str,
        parent_id: str,
    ) -> Node:
        safe_name = re.sub(r"[\\']", r"\\\g<0>", name)
        query = ' and '.join([
            f"'{parent_id}' in parents",
            f"name = '{safe_name}'",
            f'trashed = false',
        ])
        fields = f'files({FILE_FIELDS})'
        try:
            rv = await self._client.files.list_(q=query, fields=fields)
        except ResponseError as e:
            if e.status == '400':
                DEBUG('wcpan.drive.google') << 'invalid query string:' << query
                raise InvalidNameError(name) from e
            if e.status == '404':
                raise ParentNotFoundError(parent_id) from e
            raise

        rv = rv.json
        files = rv['files']
        if not files:
            return None

        node = node_from_api(files[0])
        return node

    async def _fetch_node_by_id(self, node_id: str) -> Node:
        try:
            rv = await self._client.files.get(node_id, fields=FILE_FIELDS)
        except ResponseError as e:
            if e.status == '404':
                raise NodeNotFoundError(node_id) from e
            raise
        rv = rv.json
        node = node_from_api(rv)
        return node

    async def _set_node_parent_by_id(self, node: Node, parent_id: str) -> None:
        remove_parents = [_ for _ in node.parent_list if _ != parent_id]
        api = self._client.files
        await api.update(node.id_, remove_parents=remove_parents)

    async def _force_update_by_id(self, node_id: str) -> None:
        await self._client.files.update(node_id, trashed=False)

    async def _fetch_children(self, parent_id: str) -> List[Node]:
        query = ' and '.join([
            f"'{parent_id}' in parents",
        ])
        page_token = None
        files = []
        try:
            while True:
                rv = await self._client.files.list_(
                    q=query,
                    fields=LIST_FIELDS,
                    page_size=1000,
                    page_token=page_token,
                )
                rv = rv.json
                files.extend(rv['files'])
                page_token = rv.get('nextPageToken', None)
                if not page_token:
                    break
        except ResponseError as e:
            if e.status == '404':
                raise ParentNotFoundError(parent_id) from e
            raise

        node_list = [node_from_api(_) for _ in files]
        return node_list


class GoogleReadableFile(ReadableFile):

    def __init__(self, download: Any, node: Node) -> None:
        self._download = download
        self._node = node
        self._offset = 0
        self._response = None
        self._rsps = None

    async def __aenter__(self) -> ReadableFile:
        return self

    async def __aexit__(self, type_, exc, tb) -> bool:
        await self._close_response()

    async def __aiter__(self) -> AsyncIterator[bytes]:
        # no need to make request for an empty file
        if self._node.size <= 0:
            return

        await self._open_response()
        async for chunk in self._response.chunks():
            yield chunk

    async def read(self, length: int) -> bytes:
        # nothing to read from an empty file
        if self._node.size <= 0:
            return b''

        await self._open_response()
        return await self._response.read(length)

    async def seek(self, offset: int) -> None:
        # nop for seeking an empty file
        if self._node.size <= 0:
            return

        self._offset = offset
        await self._close_response()
        await self._open_response()

    async def node(self) -> Node:
        return self._node

    async def _download_from_offset(self) -> 'aiohttp.StreamResponse':
        try:
            return await self._download(
                file_id=self._node.id_,
                range_=(self._offset, self._node.size),
                acknowledge_abuse=False,
            )
        except DownloadAbusiveFileError:
            # FIXME automatically accept abuse files for now
            WARNING('wcpan.drive.google') << f'{self._node.id_} is an abusive file'
            return await self._download(
                file_id=self._node.id_,
                range_=(self._offset, self._node.size),
                acknowledge_abuse=True,
            )

    async def _open_response(self) -> None:
        if not self._response:
            async with contextlib.AsyncExitStack() as stack:
                self._response = await stack.enter_async_context(
                    await self._download_from_offset(),
                )
                self._rsps = stack.pop_all()

    async def _close_response(self) -> None:
        if self._response:
            await self._rsps.aclose()
            self._response = None
            self._rsps = None


class GoogleWritableFile(WritableFile):

    def __init__(self,
        initiate: Any,
        upload: Any,
        get_status: Any,
        get_node: Any,
        touch: Any,
        parent_id: str,
        name: str,
        timeout: float,
        size: int = None,
        mime_type: str = None,
        media_info: MediaInfo = None,
        private: PrivateDict = None,
    ) -> None:
        self._initiate = initiate
        self._upload = upload
        self._get_status = get_status
        self._get_node = get_node
        self._touch = touch
        self._parent_id = parent_id
        self._name = name
        self._private = private
        self._timeout = timeout
        self._size = size
        self._mime_type = mime_type
        self._media_info = media_info
        self._url = None
        self._offset = None
        self._queue = asyncio.Queue(maxsize=1)
        self._bg = None
        self._rv = None

    async def __aenter__(self) -> WritableFile:
        if self._size > 0:
            self._url = await self._get_session_url()
            self._bg = None
        else:
            self._url = None
            self._bg = self._touch_empty()
        self._offset = 0
        self._queue = asyncio.Queue(maxsize=1)
        self._rv = None

        return self

    async def __aexit__(self, type_, exc, tb) -> bool:
        if self._bg:
            rv = await self._bg
            self._rv = rv.json
        else:
            # error happened before the first write
            self._rv = None
        self._bg = None
        self._queue = None
        self._offset = None
        self._url = None

    async def tell(self) -> int:
        return await self._get_offset()

    async def seek(self, offset: int) -> None:
        self._offset = offset
        await self._close_request()
        await self._open_request()

    async def write(self, chunk: bytes) -> int:
        await self._open_request()
        feed = self._queue.put(chunk)
        feed = asyncio.wait_for(feed, timeout=self._timeout)
        feed = asyncio.create_task(feed)
        await asyncio.wait([feed, self._bg],
                           return_when=asyncio.FIRST_COMPLETED)
        if not feed.done():
            # background connection ended
            feed.cancel()
        # consume exception
        feed.result()
        return len(chunk)

    async def node(self) -> Optional[Node]:
        if not self._rv:
            return None
        node = await self._get_node(self._rv['id'])
        return node

    async def _close_request(self) -> None:
        if not self._bg:
            return None
        if not self._bg.done():
            self._bg.cancel()
        try:
            rv = await self._bg
            return rv
        except (Exception, asyncio.CancelledError) as e:
            EXCEPTION('wcpan.drive.google', e) << 'close'
        finally:
            self._queue = asyncio.Queue(maxsize=1)
            self._bg = None

    async def _open_request(self) -> None:
        if self._bg:
            return
        f = self._upload_to()
        self._bg = asyncio.create_task(f)

    async def _produce(self) -> AsyncGenerator[bytes, None]:
        while True:
            async with self._get_one_chunk() as chunk:
                if not chunk:
                    break
                yield chunk

    @contextlib.asynccontextmanager
    async def _get_one_chunk(self) -> AsyncGenerator[bytes, None]:
        chunk = await self._queue.get()
        try:
            yield chunk
        finally:
            self._queue.task_done()

    async def _get_session_url(self) -> str:
        rv = await self._initiate(
            file_name=self._name,
            total_file_size=self._size,
            parent_id=self._parent_id,
            mime_type=self._mime_type,
            media_info=self._media_info,
            app_properties=self._private,
        )
        url = rv.get_header('Location')
        return url

    async def _upload_to(self) -> 'aiohttp.ClientResponse':
        try:
            rv = await self._upload(
                self._url,
                producer=self._produce,
                offset=self._offset,
                total_file_size=self._size,
                mime_type=self._mime_type,
            )
        except ResponseError as e:
            if e.status == '404':
                raise UploadError('the upload session has been expired') from e
            raise
        return rv

    async def _get_offset(self) -> int:
        try:
            rv = await self._get_status(self._url, self._size)
        except ResponseError as e:
            if e.status == '410':
                # This means the temporary URL has been cleaned up by Google
                # Drive, so the client has to start over again.
                raise UploadError('the upload session has been expired') from e
            raise

        if rv.status != '308':
            raise UploadError(f'invalid upload status: {rv.status}')
        
        try:
            rv = rv.get_header('Range')
        except KeyError:
            # no data yet
            return 0

        rv = re.match(r'bytes=0-(\d+)', rv)
        if not rv:
            raise UploadError('invalid upload range')
        rv = int(rv.group(1))
        return rv

    async def _touch_empty(self) -> 'aiohttp.ClientResponse':
        rv = await self._touch(
            self._name,
            parent_id=self._parent_id,
            mime_type=self._mime_type,
            app_properties=self._private,
        )
        return rv


class PicklableHasher(Hasher):

    def __init__(self):
        self._hasher = hashlib.md5()

    def __getstate__(self):
        return hashlib.md5

    def __setstate__(self, hasher_class):
        self._hasher = hasher_class()

    def update(self, data):
        self._hasher.update(data)

    def hexdigest(self):
        return self._hasher.hexdigest()

    def digest(self):
        return self._hasher.digest()

    def copy(self):
        return self._hasher.copy()


def normalize_changes(
    change_list: List[GoogleFileDict],
) -> Generator[ChangeDict, None, None]:
    for change in change_list:
        is_removed = change['removed']
        if is_removed:
            yield {
                'removed': True,
                'id': change['fileId'],
            }
            continue

        file_ = change['file']
        is_shared = file_['shared']
        is_owned_by_me = file_['ownedByMe']
        if is_shared or not is_owned_by_me:
            continue

        file_ = dict_from_api(file_)
        yield {
            'removed': False,
            'node': file_,
        }


def dict_from_api(data: GoogleFileDict) -> NodeDict:
    id_ = data['id']

    is_folder = data['mimeType'] == FOLDER_MIME_TYPE

    size = data.get('size', None)
    if size is not None:
        size = int(size)

    private = data.get('appProperties', None)

    image = None
    if private and 'image' in private:
        width, height = private['image'].split(' ')
        image = {
            'width': int(width),
            'height': int(height),
        }
        del private['image']
    if not image and 'imageMediaMetadata' in data:
        image = {
            'width': data['imageMediaMetadata']['width'],
            'height': data['imageMediaMetadata']['height'],
        }

    video = None
    if private and 'video' in private:
        width, height, ms_duration = private['video'].split(' ')
        video = {
            'width': int(width),
            'height': int(height),
            'ms_duration': int(ms_duration),
        }
        del private['video']
    if not video and 'videoMediaMetadata' in data:
        video = {
            'width': data['videoMediaMetadata']['width'],
            'height': data['videoMediaMetadata']['height'],
            'ms_duration': data['videoMediaMetadata']['durationMillis'],
        }

    return {
        'id': id_,
        'name': data['name'],
        'trashed': data['trashed'],
        'created': data['createdTime'],
        'modified': data['modifiedTime'],
        'parent_list': data.get('parents', None),
        'is_folder': is_folder,
        'mime_type': None if is_folder else data['mimeType'],
        'hash': data.get('md5Checksum', None),
        'size': size,
        'image': image,
        'video': video,
        'private': private,
    }


def node_from_api(data: GoogleFileDict) -> Node:
    node_data = dict_from_api(data)
    return Node.from_dict(node_data)



import contextlib
import json
from typing import List, Tuple, Dict

import arrow
from wcpan.drive.core.types import MediaInfo

from .network import Network, Response, ContentProducer
from .util import FOLDER_MIME_TYPE, OAuth2Storage


API_ROOT = 'https://www.googleapis.com/drive/v3'


class Client(object):

    def __init__(self, storage: OAuth2Storage, timeout: int) -> None:
        self._storage = storage
        self._timeout = timeout
        self._network = None
        self._api = None
        self._raii = None

    async def __aenter__(self) -> 'Client':
        async with contextlib.AsyncExitStack() as stack:
            self._network = await stack.enter_async_context(
                Network(self._storage, self._timeout))
            self._api = {
                'changes': Changes(self._network),
                'files': Files(self._network),
            }
            self._raii = stack.pop_all()

        return self

    async def __aexit__(self, type_, value, traceback) -> bool:
        await self._raii.aclose()
        self._api = None
        self._network = None
        self._raii = None

    @property
    def changes(self) -> 'Changes':
        return self._api['changes']

    @property
    def files(self) -> 'Files':
        return self._api['files']


class Changes(object):

    def __init__(self, network: Network) -> None:
        self._network = network
        self._root = API_ROOT + '/changes'

    async def get_start_page_token(self,
        *,
        drive_id: str = None,
        fields: str = None,
    ) -> Response:
        args = {}
        if drive_id is not None:
            args['driveId'] = drive_id
        if fields is not None:
            args['fields'] = fields

        uri = self._root + '/startPageToken'
        rv = await self._network.fetch('GET', uri, args)
        return rv

    async def list_(self,
        page_token: str,
        *,
        drive_id: str = None,
        fields: str = None,
        include_corpus_removals: bool = None,
        include_removed: bool = None,
        page_size: int = None,
        restrict_to_my_drive: bool = None,
        spaces: str = None,
    ) -> Response:
        args = {
            'pageToken': page_token,
        }
        if drive_id is not None:
            args['driveId'] = drive_id
        if fields is not None:
            args['fields'] = fields
        if include_corpus_removals is not None:
            args['includeCorpusRemovals'] = include_corpus_removals
        if include_removed is not None:
            args['includeRemoved'] = include_removed
        if page_size is not None:
            args['pageSize'] = page_size
        if restrict_to_my_drive is not None:
            args['restrictToMyDrive'] = restrict_to_my_drive
        if spaces is not None:
            args['spaces'] = spaces

        rv = await self._network.fetch('GET', self._root, args)
        return rv


class Files(object):

    def __init__(self, network: Network) -> None:
        self._network = network
        self._root = API_ROOT + '/files'
        self._upload_uri = 'https://www.googleapis.com/upload/drive/v3/files'

    # only for metadata
    async def get(self,
        file_id: str,
        *,
        fields: str = None,
    ) -> Response:
        args = {}
        if fields is not None:
            args['fields'] = fields

        uri = self._root + '/' + file_id
        rv = await self._network.fetch('GET', uri, args)
        return rv

    async def list_(self,
        *,
        corpora: str = None,
        drive_id: str = None,
        fields: str = None,
        order_by: str = None,
        page_size: int = None,
        page_token: str = None,
        q: str = None,
        spaces: str = None,
    ) -> Response:
        args = {}
        if corpora is not None:
            args['corpora'] = corpora
        if drive_id is not None:
            args['driveId'] = drive_id
        if fields is not None:
            args['fields'] = fields
        if order_by is not None:
            args['orderBy'] = order_by
        if page_size is not None:
            args['pageSize'] = page_size
        if page_token is not None:
            args['pageToken'] = page_token
        if q is not None:
            args['q'] = q
        if spaces is not None:
            args['spaces'] = spaces

        rv = await self._network.fetch('GET', self._root, args)
        return rv

    async def download(self,
        file_id: str,
        range_: Tuple[int, int],
        *,
        acknowledge_abuse: bool = None,
    ) -> Response:
        args = {
            'alt': 'media',
        }
        if acknowledge_abuse is not None:
            args['acknowledgeAbuse'] = acknowledge_abuse

        headers = {
            'Range': f'bytes={range_[0]}-{range_[1]}',
        }

        uri = self._root + '/' + file_id
        rv = await self._network.download('GET', uri, args, headers=headers)
        return rv

    async def initiate_uploading(self,
        file_name: str,
        total_file_size: int,
        *,
        parent_id: str = None,
        mime_type: str = None,
        media_info: MediaInfo = None,
        app_properties: Dict[str, str] = None,
    ) -> Response:
        if not file_name:
            raise ValueError('file name is empty')
        if total_file_size <= 0:
            raise ValueError('please use create_empty_file() to create an empty file')

        metadata = {
            'name': file_name,
        }
        if parent_id is not None:
            metadata['parents'] = [parent_id]

        props = {}
        if app_properties:
            props.update(app_properties)
        if media_info and media_info.is_image:
            props['image'] = f'{media_info.width} {media_info.height}'
        if media_info and media_info.is_video:
            props['video'] = f'{media_info.width} {media_info.height} {media_info.ms_duration}'
        if props:
            metadata['appProperties'] = props

        metadata = json.dumps(metadata)
        metadata = metadata.encode('utf-8')
        headers = {
            'X-Upload-Content-Length': total_file_size,
            'Content-Type': 'application/json; charset=UTF-8',
            'Content-Length': len(metadata),
        }
        if mime_type is not None:
            headers['X-Upload-Content-Type'] = mime_type

        args = {
            'uploadType': 'resumable',
        }

        rv = await self._network.fetch('POST', self._upload_uri, args,
                                       headers=headers, body=metadata)
        return rv

    async def upload(self,
        uri: str,
        producer: ContentProducer,
        offset: int,
        total_file_size: int,
        *,
        mime_type: str = None,
    ) -> Response:
        if not uri:
            raise ValueError('invalid session URI')
        if not producer:
            raise ValueError('invalid body producer')
        if total_file_size <= 0:
            raise ValueError('please use create_empty_file() to create an empty file')
        if offset < 0 or offset >= total_file_size:
            raise ValueError('offset is out of range')

        last_position = total_file_size - 1
        headers = {
            'Content-Length': total_file_size - offset,
            'Content-Range': f'bytes {offset}-{last_position}/{total_file_size}',
        }
        if mime_type is not None:
            headers['Content-Type'] = mime_type

        # Producer may raise timeout error, not the upload itself.
        rv = await self._network.upload('PUT', uri, headers=headers,
                                        body=producer)
        return rv

    async def get_upload_status(self,
        uri: str,
        total_file_size: int,
    ) -> Response:
        if not uri:
            raise ValueError('invalid session URI')
        if total_file_size <= 0:
            raise ValueError('please use create_empty_file() to create an empty file')

        headers = {
            'Content-Length': 0,
            'Content-Range': f'bytes */{total_file_size}',
        }
        rv = await self._network.fetch('PUT', uri, headers=headers)
        return rv

    async def create_folder(self,
        folder_name: str,
        *,
        parent_id: str = None,
        app_properties: Dict[str, str] = None,
    ) -> Response:
        metadata = {
            'name': folder_name,
            'mimeType': FOLDER_MIME_TYPE,
        }
        if parent_id is not None:
            metadata['parents'] = [parent_id]
        if app_properties:
            metadata['appProperties'] = app_properties
        metadata = json.dumps(metadata)
        metadata = metadata.encode('utf-8')
        headers = {
            'Content-Type': 'application/json; charset=UTF-8',
            'Content-Length': len(metadata),
        }

        rv = await self._network.fetch('POST', self._root, headers=headers,
                                       body=metadata)
        return rv

    async def create_empty_file(self,
        file_name: str,
        *,
        parent_id: str = None,
        mime_type: str = None,
        app_properties: Dict[str, str] = None,
    ) -> Response:
        if not file_name:
            raise ValueError('file name is empty')

        metadata = {
            'name': file_name,
        }
        if parent_id is not None:
            metadata['parents'] = [parent_id]
        if mime_type is not None:
            metadata['mimeType'] = mime_type
        else:
            metadata['mimeType'] = 'application/octet-stream'
        if app_properties:
            metadata['appProperties'] = app_properties
        metadata = json.dumps(metadata)
        metadata = metadata.encode('utf-8')
        headers = {
            'Content-Type': 'application/json; charset=UTF-8',
            'Content-Length': len(metadata),
        }

        rv = await self._network.fetch('POST', self._root, headers=headers,
                                       body=metadata)
        return rv

    async def update(self,
        file_id: str,
        *,
        add_parents: List[str] = None,
        remove_parents: List[str] = None,
        name: str = None,
        trashed: bool = None,
        app_properties: Dict[str, str] = None,
        media_info: MediaInfo = None,
    ) -> Response:
        args = {}
        if add_parents is not None:
            args['addParents'] = ','.join(add_parents)
        if remove_parents is not None:
            args['removeParents'] = ','.join(remove_parents)

        metadata = {}
        if name is not None:
            metadata['name'] = name
        if trashed is not None:
            metadata['trashed'] = trashed

        props = {}
        if app_properties:
            props.update(app_properties)
        if media_info and media_info.is_image:
            props['image'] = f'{media_info.width} {media_info.height}'
        if media_info and media_info.is_video:
            props['video'] = f'{media_info.width} {media_info.height} {media_info.ms_duration}'
        if props:
            metadata['appProperties'] = props

        if not args and not metadata:
            raise ValueError('not enough parameter')

        now = arrow.utcnow()
        metadata['modifiedTime'] = now.isoformat()

        metadata = json.dumps(metadata)
        metadata = metadata.encode('utf-8')
        headers = {
            'Content-Type': 'application/json; charset=UTF-8',
            'Content-Length': len(metadata),
        }

        uri = self._root + '/' + file_id
        rv = await self._network.fetch('PATCH', uri, args, headers=headers,
                                       body=metadata)
        return rv



from typing import TypedDict, Optional, Dict, Any, List
import asyncio
import contextlib
import json
import pathlib
import urllib.parse

import aiohttp
import yaml

from wcpan.logger import DEBUG, EXCEPTION

from .exceptions import AuthenticationError, CredentialFileError, TokenFileError


class OAuth2Info(TypedDict):

    client_id: str
    client_secret: str
    redirect_uri: str
    access_token: str
    refresh_token: str


FOLDER_MIME_TYPE = 'application/vnd.google-apps.folder'
OAUTH_TOKEN_VERSION = 1


class OAuth2Storage(object):

    def __init__(self,
        config_path: pathlib.Path,
        data_path: pathlib.Path,
    ) -> None:
        self._client_secret = config_path / 'client_secret.json'
        self._oauth_token = data_path / 'oauth_token.yaml'

    def load_oauth2_info(self) -> OAuth2Info:
        # load API key
        with self._client_secret.open('r') as fin:
            client = json.load(fin)
        try:
            client = client['installed']
            redirect_uri = client['redirect_uris'][0]
            client_id = client['client_id']
            client_secret = client['client_secret']
        except (KeyError, IndexError):
            raise CredentialFileError()

        # load refresh token
        if not self._oauth_token.is_file():
            access_token = None
            refresh_token = None
        else:
            with self._oauth_token.open('r') as fin:
                token = yaml.safe_load(fin)
            version = token.get('version', 0)
            if version != OAUTH_TOKEN_VERSION:
                raise TokenFileError(f'invalid token version: {version}')
            try:
                access_token = token['access_token']
                refresh_token = token['refresh_token']
            except KeyError:
                raise TokenFileError(f'invalid token format')

        return {
            'client_id': client_id,
            'client_secret': client_secret,
            'redirect_uri': redirect_uri,
            'access_token': access_token,
            'refresh_token': refresh_token,
        }

    def save_oauth2_info(self,
        access_token: str,
        refresh_token: str,
    ) -> None:
        token = {
            'version': OAUTH_TOKEN_VERSION,
            'access_token': access_token,
            'refresh_token': refresh_token,
        }

        # save refresh token
        with self._oauth_token.open('w') as fout:
            yaml.dump(token, fout, default_flow_style=False)


class OAuth2Manager(object):

    def __init__(self,
        session: aiohttp.ClientSession,
        storage: OAuth2Storage,
    ) -> None:
        self._session = session
        self._storage = storage
        self._lock = asyncio.Condition()
        self._refreshing = False
        self._error = False
        self._oauth = None
        self._raii = None

    async def __aenter__(self) -> 'OAuth2Manager':
        oauth2_info = self._storage.load_oauth2_info()

        async with contextlib.AsyncExitStack() as stack:
            self._oauth = await stack.enter_async_context(
                OAuth2CommandLineAuthenticator(
                    self._session,
                    oauth2_info['client_id'],
                    oauth2_info['client_secret'],
                    oauth2_info['redirect_uri'],
                    oauth2_info['access_token'],
                    oauth2_info['refresh_token'],
                ))
            self._storage.save_oauth2_info(
                self._oauth.access_token,
                self._oauth.refresh_token,
            )
            self._raii = stack.pop_all()

        return self

    async def __aexit__(self, type_, exc, tb) -> bool:
        await self._raii.aclose()
        self._raii = None
        self._oauth = None
        self._error = False
        self._refreshing = False

    async def get_access_token(self) -> str:
        if self._refreshing:
            async with self._lock:
                await self._lock.wait()
        if self._error:
            raise AuthenticationError()
        return self._oauth.access_token

    async def renew_token(self):
        if self._refreshing:
            async with self._lock:
                await self._lock.wait()
            return

        async with self._guard():
            try:
                await self._oauth.refresh()
                self._storage.save_oauth2_info(
                    self._oauth.access_token,
                    self._oauth.refresh_token,
                )
            except Exception as e:
                EXCEPTION('wcpan.drive.google', e) << 'error on refresh token'
                self._error = True
                raise
            self._error = False

        DEBUG('wcpan.drive.google') << 'refresh access token'

    @contextlib.asynccontextmanager
    async def _guard(self):
        self._refreshing = True
        try:
            yield
        finally:
            self._refreshing = False
            async with self._lock:
                self._lock.notify_all()


class OAuth2CommandLineAuthenticator(object):

    def __init__(self,
        session: aiohttp.ClientSession,
        client_id: str,
        client_secret: str,
        redirect_uri: str,
        access_token: str = None,
        refresh_token: str = None
    ) -> None:
        self._session = session
        self._client_id = client_id
        self._client_secret = client_secret
        self._redirect_uri = redirect_uri
        self._access_token = access_token
        self._refresh_token = refresh_token

    async def __aenter__(self) -> 'OAuth2CommandLineAuthenticator':
        if self._access_token is None:
            await self._fetch_access_token()
        return self

    async def __aexit__(self, type_, value, traceback) -> bool:
        pass

    @property
    def access_token(self) -> str:
        assert self._access_token is not None
        return self._access_token

    @property
    def refresh_token(self) -> Optional[str]:
        return self._refresh_token

    async def refresh(self) -> None:
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        body = urllib.parse.urlencode({
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'refresh_token': self._refresh_token,
            'grant_type': 'refresh_token',
        })

        async with self._session.post(self.oauth_access_token_url,
                                      headers=headers, data=body) as response:
            response.raise_for_status()
            token = await response.json()
        self._save_token(token)

    def _save_token(self, token: Dict[str, Any]) -> None:
        self._access_token = token['access_token']
        if 'refresh_token' in token:
            self._refresh_token = token['refresh_token']

    async def _fetch_access_token(self) -> None:
        # get code on success
        code = await self._authorize_redirect()
        token = await self._get_authenticated_user(code=code)
        self._save_token(token)

    async def _authorize_redirect(self) -> str:
        kwargs = {
            'redirect_uri': self._redirect_uri,
            'client_id': self._client_id,
            'response_type': 'code',
            'scope': ' '.join(self.scopes),
        }

        url = urllib.parse.urlparse(self.oauth_authorize_url)
        url = urllib.parse.urlunparse((
            url[0],
            url[1],
            url[2],
            url[3],
            urllib.parse.urlencode(kwargs),
            url[5],
        ))
        return await self.redirect(url)

    @property
    def oauth_authorize_url(self) -> str:
        return 'https://accounts.google.com/o/oauth2/auth'

    @property
    def oauth_access_token_url(self) -> str:
        return 'https://accounts.google.com/o/oauth2/token'

    @property
    def scopes(self) -> List[str]:
        return [
            'https://www.googleapis.com/auth/drive',
        ]

    async def _get_authenticated_user(self, code: str) -> Dict[str, Any]:
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        body = urllib.parse.urlencode({
            'redirect_uri': self._redirect_uri,
            'code': code,
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'grant_type': 'authorization_code',
        })
        async with self._session.post(self.oauth_access_token_url,
                                      headers=headers, data=body) as response:
            response.raise_for_status()
            return await response.json()

    # NOTE Use case depends
    async def redirect(self, url: str) -> str:
        print(url)
        return input().strip()
